%	MATLAB program: efrit.m, version 2.0
%	
%	<< E-FRIT (Extended Fictitious Reference Iterative Tuning) >>
%	
%	The E-FRIT is a direct controller tuning algorithm, which can 
%	optimize control parameters from closed-loop operation data 
%	without building a process model. 
%	The E-FRIT algorithm was developd by Workshop No.27, which was 
%	supported by JSPS PSE 143rd committee, in order to solve the 
%	problem that process control engineers had encountered in the 
%	process industry: how to tune PID parameters without modeling.
%	Here is an asnwer.
%	So far, E-FRIT has been used in several petrochemical companies 
%	and has succeeded in improving the control performance and  
%	increasing profits.
%	We hope you can benefit from this program of E-FRIT.
%	
%	For more information on JSPS PSE 143 Workshop No.27 and 
%	the latest information on E-FRIT, check
%	http://ws27.pse143.org/
%	http://e-frit.chase-dream.com/
%	
%		Manabu Kano and Morimasa Ogawa, May 22, 2010
%	
%	--- Japanese ---
%	E-FRIT́CvZXf\z邱ƂȂC[v̉^]f[
%	^œKȐp[^𓱏ołPID@łDE-FRIT͓{
%	wpUvZXVXeHwPSRψ[NVbvNo.27
%	JC܂łɕ̊ƂŐ䐫\PƌoϓIvɍv
%	Ă܂DE-FRITvOpD
%	
%	{wpUvZXVXeHw143ψ[NVbvNo.27
%	E-FRITɊւŐVɂĂ͉LTCgmFD
%	http://ws27.pse143.org/
%	http://e-frit.chase-dream.com/
%	
%		2010N522@[ wC 琳
%	
%	--- update history ---
%	EFRIT version 0.0
%		original program
%		(02/10, 2009) K. Tasaka and M. Kano
%	EFRIT version 0.1
%		(03/31, 2010) M. Ogawa
%		>> use NLP instead of PSO for faster optimization.
%	EFRIT version 0.2
%		(04/10, 2010) M. Ogawa and M. Kano
%	EFRIT version 0.3
%		(04/13, 2010) M. Ogawa and M. Kano
%		>> modify scaling factor calc.
%	EFRIT version 1.0
%		(05/22, 2010) M. Ogawa and M. Kano
%		>> minor modification for the release
%	EFRIT version 1.1
%		(05/27, 2010) M. Ogawa and M. Kano
%		>> add PID
%	EFRIT version 2.0
%		(10/04, 2010) M. Ogawa
%		>> enable to set upper bound on Td/Ti
%		>> modify scaling factor on du; 
%		   lambda=1.0 is suggested for any algorithm
%
%	Copyright (C) 2009-2010 Kenichi Tasaka, Manabu Kano, Morimasa Ogawa.
%	All rights reserved. 
%	{vOі{vOύXvO̍Ĕzzւ܂D
%
	clear all;
	clc;

	global  y0  u0  r0  t0
	global  PID_algorithm  dir_rev  gamma
	global  n  Tn
	global  fs  lambda
	global  rt  yt
%
% ---------------------------------------------------------
% --- load closed-loop test data under set-point change ---
% ---------------------------------------------------------
%
%	PID_algorithm:
%		1 = PID control
%		2 = PI-D control
%		3 = I-PD control
%	dir_rev: control direction to ensure negative feedback 
%		 1 = direct : e (+) -> u (+)
%		-1 = reverse: e (+) -> u (-)
%
%	Kc0		: proportional gain		(%/%)
%	Ti0		: integral (reset) time	(min)
%	Td0		: derivative time		(min)
%	gamma	: derivative gain		(-)
%
%	tau		: sampling interval		(min)
%	rs		: set point				(%)
%	ys		: controlled variable	(%)
%	us		: manipulated variable	(%)
%
	CL_test_case =...
	input('case of closed-loop test ( 0 = Actual Problem, 1 = LC, 2 = AC, 3 = TC, 4 = FC ) = ');

	switch CL_test_case
	case 0; % acual problem defined by user
		load CL_Test_Data;
	case 1;	% solvent column base level control by reboiler steam
		load CL_Test_Data_LC;
	case 2;	% analyzer feedback control with fast process dynamics
		load CL_Test_Data_AC;
	case 3;	% reaction temp. control of polmerization reactor
		load CL_Test_Data_TC;
	case 4;	% reboiler steam flow control by manipulating valve on condensate
		load CL_Test_Data_FC;
	otherwise
		disp('unknown test case');
	end;

	kmax = length(rs);
	ts   = 0:tau:(kmax-1)*tau;
%
% ------------------------------------------------------
% --- smoothen ys and us using Hamming-window filter ---
% ------------------------------------------------------
%
% --- low pass FIR filter ---
%
	nf = 10;			% FIR order
	if CL_test_case == 3
		fc = 0.10;		% FIR filter cut-off frequency 1 = Nyquist freq.
	else
		fc = 0.50;		% FIR filter cut-off frequency 1 = Nyquist freq.
	end	
	b  = fir1(nf,fc);	% FIR series
%
% --- smoothing ---
%
	nf2 = ceil( nf / 2 );
	k1  = 1 + nf2;
	k2  = kmax - nf2;

	rf = rs(k1:k2);		% no filtering on rs
	kmax = length(rf);
	t  = ( 0:tau:(kmax-1)*tau )';
	uf = zeros(kmax,1);
	yf = zeros(kmax,1);

	kk = 0;				% yf & uf pointer
	for k = k1:+1:k2
		kw1 = k - nf2;
		kw2 = k + nf2;
		uw  = us(kw1:kw2);
		yw  = ys(kw1:kw2);
		kk  = kk + 1;
		uf(kk) = b * uw;
		yf(kk) = b * yw;
	end;
%
% -------------------------------
% --- input tuning parameters ---
% -------------------------------
%
% --- n, T99: order & time constant of reference model 
%     to obtain desired settling time for set-point change ---
% --- lambda: weight coefficient on delta-u ---
%
	switch PID_algorithm;
	case 1;	% PID algorithm
		n = 2;
	case 2;	% PI-D algorithm
		n = 2;
	case 3;	% I-PD algorithm
		n = 3;
	otherwise
		disp('unknown PID algorithm');
	end
	
	switch CL_test_case
	case 0;	% actual problem
		disp('input desired settling time');
	case 1;	% solvent column base level control by reboiler steam
		disp('suggested T99 = 40 min');
	case 2;	% analyzer control with fast process dynamics
		disp('suggested T99 = 4.0 min');
	case 3;	% reaction temp. control of polmerization reactor
		disp('suggested T99 = 10 min');
	case 4;	% reboiler steam flow control by manipulating valve on condensate
		disp('suggested T99 = 15 min');
	otherwise
		disp('unknown test case');
	end

	T99 = input('99% response time, T99 (min) = ');
	Tn  = T99 / ( 4.4 * n^0.6 );	% time constant of reference model
	lambda = input('lambda (suggested 1.0) = ');
%
% --- alpha: upper limit of Td / Ti ---
%
	alpha = input('upper bound of Td/Ti (<=0.2 to use derivative action / =0 without derivative action) = ');
%
% -----------------------------
% --- estimate PID settings ---
% -----------------------------
%
% --- remove bias ---
%
	r0 = rf - rf(1);
	y0 = yf - rf(1);
	u0 = uf - uf(1);
	t0 = t;
%
% --- starting guess of PID parameters = PID settings at closed-loop test ---
%
	TL0 = 0;				% dead time	(min)
	theta0 = [ Kc0, Ti0, Td0, TL0 ];
%
% --- C and F ---
%
	num_C = dir_rev * Kc0 * [ ( 1 + ( 1 / gamma ) ) * Ti0 * Td0  ( Ti0 + ( Td0 / gamma ) )  1 ];
	den_C = [  ( Ti0 * Td0 / gamma )  Ti0  0 ];
	C     = tf( num_C,den_C );
	Cinv  = tf( den_C,num_C );

	switch PID_algorithm;
	case 1; % PID algorithm
		num_F = [ 1 ];
		den_F = [ 1 ];
		F = tf( num_F,den_F );
	case 2; % PI-D algorithm
		num_F = [ Ti0 * Td0 / gamma   ( Ti0 + Td0 / gamma )   1 ];
		den_F = [ ( 1 + ( 1 / gamma ) ) * Ti0 * Td0   ( Ti0 + Td0 / gamma )   1 ];
		F = tf( num_F,den_F );
	case 3; % I-PD algorithm
		num_F = [ Td0 / gamma   1 ];
		den_F = [ ( 1 + ( 1 / gamma ) ) * Ti0 * Td0   ( Ti0 + Td0 / gamma )   1 ];
		F = tf( num_F,den_F );
	otherwise
		disp('unknown PID algorithm');
	end
%
% --- fs: scaling factor on delta_u^2 ---
%
	M   = reference_model(Tn,TL0,n);

	rt  = lsim(Cinv,u0,t0) + y0;
	yt  = lsim( M/F,rt,t0 );
	e   = y0 - yt;
	Je  = ( e' *  e ) / kmax;

	et0  = rt - yt;
	ut0  = lsim( C,et0,t0 );
	dut0 = ut0(1:kmax) - [ 0; ut0(1:kmax-1) ];
	Ju   = ( dut0' * dut0 ) / kmax;

	fs = sqrt( Je / Ju );		% scaling factor

	% version 1.1 or earlier
	% yt0 = lsim( M,r0,t0 );	% desired response
	% e   = yt0 - y0;
	% Je  = ( e' *  e ) / kmax;
	% du  = u0(1:kmax) - [ 0; u0(1:kmax-1) ];
	% Ju  = ( du' * du ) / kmax;
	% fs = sqrt( Je / Ju );	% scaling factor
%
% --- lower / upper limit of parameters ---
%
%	        Kc   Ti   Td   TL
%	       %/%  min  min  min
	ub = [  50, 150,  30,  10 ];	% upper bound
	lb = [ 0.1, 0.1,   0,   0 ];	% lower bound for theta
%
% --- constraint Td/Ti <= alpha ( -alpha * theta(2) + 1 * theta(3) ) <= 0 ---
%
	A = [ 0  -alpha  1  0 ];
	b = 0;
%
% --- optimization ---
%
	options = [];
	options = optimset('LargeScale','on','Algorithm','active-set','DerivativeCheck','on');

	time_start = clock;		% get start time
	[ theta, fval, exitflag, output] = fmincon('objfun',theta0,A,b,[],[],lb,ub,'funcon',options);
	time_stop = clock;		% get converged time
%
% --- results ---
%
%	output optimization information
	disp('convergence time (sec)')
	c_time = etime(time_stop,time_start)
						% convergence time	(sec)

	disp('PID settings: Kc (%/%), Ti (min), Td (min)')
	Kc = theta(1)		% proportional gain	(%/%)
	Ti = theta(2)		% reset time		(min)
	Td = theta(3)		% derivative time	(min)

	disp('process dead time: TL (min)')
	TL = theta(4)		% dead time		(min)
%
% ------------
% --- plot ---
% ------------
%
% --- raw and filtered data ---
%
	tp = ts(k1:k2) - ts(k1);	% plot time

	figure(1);clf;
	subplot(211);plot(t,rf,'k-',tp,ys(k1:k2),'b-',t,yf,'r-');grid;
	title('set-point (black); controlled variable raw (blue), filtered (red)');
	ylabel('PV (%)');xlabel('time (min)');

	subplot(212);plot(tp,us(k1:k2),'b-',t,uf,'r-');grid;
	title('manipulated variable raw (blue), filtered (red)');
	ylabel('MV (%)');xlabel('time (min)');
%
% --- results ---
%
	r0 = r0 + rf(1);
	y0 = y0 + rf(1);
	u0 = u0 + uf(1);
	yt = yt + rf(1);
	rt = rt + rf(1);

	figure(2);clf;
	subplot(311);plot(t,r0,'b-',t,rt,'r-');grid;
	title('tested set-point (blue); fictitious set-point (red)');
	ylabel('SV (%)');

	subplot(312);plot(t,y0,'b-',t,yt,'r-');grid;
	title('tested controlled variable (blue); desired response (red)');
	ylabel('PV (%)');

	subplot(313);plot(t,u0,'b-');grid;
	title('tested manipulated variable');
	ylabel('MV (%)');xlabel('time (min)');
	